import multiprocessing
import operator
import logging
import numpy as np

from sklearn.cross_validation import cross_val_score
from sklearn.cross_validation import train_test_split

from automl.pipeline import ModelSpaceFunctor, PipelineData


class ModelSpace:
    """Class contains predefined models with given hyperparameters

    Parameters
    ----------
    model_list : list of tuples
        each element must be a tuple where the first element is an estimator
        and the second one is estimator parameter kwargs dict
    """

    def __init__(self, model_list):
        self._log = logging.getLogger(self.__class__.__name__)
        self._model_list = model_list

    def __call__(self, pipeline_data, context):
        """ Sets model space in PipelineContext and passes unchanged dataset
        on the next step

        Parametrs
        ---------
        dataset : Dataset
            Processed dataset

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        dataset : Dataset
            Unchanged dataset is passed to next step in pipeline
        """
        context.model_space = self._model_list
        return pipeline_data


class ValidationResult:
    """A wrapper class for model validation results"""

    def __init__(self, model, params, score):
        self.model_class = model.__class__
        self.model = model
        self.params = params
        self.score = score

class CV(ModelSpaceFunctor):
    """
    Class for cross-validation step in pipeline
    """
    def __init__(self, scoring, n_folds=3, n_jobs=-1):
        self._n_folds = n_folds
        self._scoring = scoring

        if n_jobs is None:
            self._n_jobs = multiprocessing.cpu_count()
        else:
            self._n_jobs = n_jobs

    def __call__(self, pipeline_data, context, hparams=None):
        model, model_params = context

        # still a bit ugly...
        # 1. model_params contain hyperopt template if we are using Hyperopt
        # 2. hparams is passsed by hyperopy only when using Hyperopt step
        if hparams is not None:
            params = hparams
        else:
            params = model_params
        model = model(**params)

        cv_score = cross_val_score(
            model,
            pipeline_data.dataset.data,
            pipeline_data.dataset.target,
            cv=self._n_folds,
            n_jobs=self._n_jobs,
            scoring=self._scoring)

        model.fit(pipeline_data.dataset.data, pipeline_data.dataset.target)

        result = ValidationResult(model, params, np.mean(cv_score))
        return result


class Validate(ModelSpaceFunctor):
    """
    Class for validation step in pipeline with using user metrics
    """

    def __init__(self, test_size, metrics, stratify=False):
        """
        Parametrs
        ---------
        test_size : float, int
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the test split.
            If int, represents the absolute number of test samples.

        metrics : callable
            Should get predicted and test vector for calculating metrics score
        """
        self._test_size = test_size
        self._metrics = metrics
        self._stratify = stratify

    def __call__(self, pipeline_data, context, hparams=None):
        """
        Executes validation for all model in model_space.

        Parametrs
        ---------
        dataset : Dataset
            Processed dataset

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        dataset : Dataset
            Unchanged dataset is passed to next step in pipeline

        validation_results : list of tuples
            Tuples like (model, score)
        """
        model, model_params = context

        # still a bit ugly...
        # 1. model_params contain hyperopt template if we are using Hyperopt
        # 2. hparams is passsed by hyperopy only when using Hyperopt step
        if hparams is not None:
            params = hparams
        else:
            params = model_params
        model = model(**params)

        stratify = None
        if self._stratify:
            stratify = pipeline_data.dataset.target

        X_train, X_test, y_train, y_test = train_test_split(
            pipeline_data.dataset.data,
            pipeline_data.dataset.target,
            test_size=self._test_size,
            stratify=stratify,
            random_state=42)

        model.fit(X_train, y_train)
        val_score = self._metrics(y_test, model.predict(X_test))
        result = ValidationResult(model, params, val_score)
        return result


class ChooseBest:
    """
    Chooses best model by scores on CV or Validation step in pipeline
    """

    def __init__(self, k, by_largest_score=True):
        """
        Parametrs
        ---------
        k : int
            Number of models for choice
        """
        self._log = logging.getLogger(self.__class__.__name__)
        self._k = k
        self._by_largest = by_largest_score

    def __call__(self, pipeline_data, context):
        """
        Parametrs
        ---------
        dataset : Dataset
            Processed dataset

        model_scorer : list of tuples
            Tuples like (model, score)

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        dataset : Dataset
            Unchanged dataset is passed to next step in pipeline

        sorted_scores : list of tuples
            Only the top self._k tuples like (model, score) with the best score
        """
        model_scores = [inp for inp in pipeline_data.return_val]
        sorted_scores = sorted(model_scores, key=operator.attrgetter('score'),
                               reverse=self._by_largest)
        self._log.info("Final model scores:")
        for model_score in sorted_scores:
            self._log.info(f"{model_score.model} - {model_score.score}")
        return PipelineData(pipeline_data.dataset, sorted_scores[:self._k])



import logging
from tqdm import tqdm
from automl.constants import *


class PipelineContext:
    """Global context of a Pipeline"""
    def __init__(self):
        self.epoch = 0
        self.prev_step = None
        self.feature_scores = None
        self.model_space = []


class PipelineData:
    def __init__(self, dataset, return_val=None):
        self.dataset = dataset
        self.return_val = return_val


class PipelineStep:
    """Base class for all pipeline steps

    Example:
        >>> PipelineStep('step name', lambda x: x * 2)
    """

    def __init__(self, name, func, initializer=False, *args, **kwargs):
        self._log = logging.getLogger(self.__class__.__name__)
        self._func = func
        self.name = name
        self._args = args
        self._kwargs = kwargs
        self._initializer = initializer
        self._cached_response = None

    def __call__(self, pipe_input, context):
        if self._initializer:
            if context.epoch == 0:
                self._initializer_was_run = True
                self._cached_response = self._func(pipe_input,
                                                   context,
                                                   *self._args,
                                                   **self._kwargs)
                return self._cached_response
            else:
                self._log.info((f"Initializer step {self.name} was already"
                                " run, skipping"))
                return self._cached_response
        else:
            return self._func(pipe_input, context, *self._args, **self._kwargs)

    def is_model_space_functor(self):
        return isinstance(self._func, ModelSpaceFunctor)


class ModelSpaceFunctor:
    """Use this class to mark any PipelineStep to be run on each
    model/parameter set from PipelineContext.model_space"""


class Pipeline:
    """AutoML Pipeline

    Example:
        >>> Pipeline() >> PipelineStep('hello step', lambda: print('Hi!'))
    """

    def __init__(self, steps=None):
        self._log = logging.getLogger(self.__class__.__name__)
        if steps is not None:
            self.steps = steps
        else:
            self.steps = []

    def __rshift__(self, other):
        if not isinstance(other, PipelineStep):
            if callable(other):
                other = PipelineStep(other.__class__.__name__, other)
            else:
                raise ValueError(("Non-callable step passted to the pipeline."
                                  f"Step {other} must be callable"))

        self.steps.append(other)
        return self

class LocalExecutor:
    """Run AutoML Pipeline locally

    Example:
    >>> LocalExecutor(epochs=10) << pipeline
    """

    def __init__(self, input_data=None, *args, **kwargs):
        """
        Parameters
        ----------
        kwargs  
            optional input parameters which will be passed to `run` method,
            useful when you launch pipeline directly via `>>` operator
        """
        self._log = logging.getLogger(self.__class__.__name__)
        self._context = PipelineContext()
        self._args = args
        self._kwargs = kwargs
        self._input_data = input_data

    def __lshift__(self, other):
        return self.run(other, self._input_data, *self._args, **self._kwargs)

    def run(self, pipeline, input_data=None, epochs=1):
        """Run pipeline.

        Parameters
        ----------
        pipeline: Pipeline
        input_data
            inuput to be passed to the pipeline
        epochs: int
            run pipeline a given number of epochs

        Returns
        -------
        result
            result of a given pipeline"""
        
        print(AUTOML_HEADER)

        self._log.info(f"Framework version: v{AUTOML_VERSION}")

        if not isinstance(input_data, PipelineData):
            input_data = PipelineData(input_data)

        for epoch_n in range(0, epochs):
            self._context.epoch = epoch_n
            self._log.info(f"Starting AutoML Epoch #{epoch_n + 1}")

            pipeline_data = input_data

            if pipeline_data.dataset:
                self._log.info(f"Dataset columns: {pipeline_data.dataset.columns}")
            for step in tqdm(pipeline.steps):
                self._log.info(f"Running step '{step.name}'")
                if step.is_model_space_functor():
                    pipeline_data.return_val = [step(pipeline_data, model_and_params)
                                                for model_and_params
                                                in self._context.model_space]
                else:
                    pipeline_data = step(pipeline_data, self._context)

        return self._context, pipeline_data



import logging
import random
from automl.pipeline import PipelineStep


class RandomChoice:
    """Combinator that chooses one of the steps uniformly at random.
    >>> pipeline >> RandomChoice([step1, step2])
    """

    def __init__(self, steps):
        """Initialize combinator.

        Parameters
        ----------
        steps: iterable of PipelineStep
           steps to choose from 
        """
        self.name = self.__class__.__name__
        self._steps = steps

    def __call__(self, data, context):
        step_choice = random.randint(0, len(self._steps) - 1)
        return self._steps[step_choice](data, context)



"""Pipeline steps and general classes for hyperparameter optimization"""

import logging
from functools import partial
from hyperopt import fmin, tpe, STATUS_OK, Trials
from automl.pipeline import ModelSpaceFunctor


class HyperparameterSearchResult:
    """A container for the results of a parameter seach"""

    def __init__(self, best_model, best_score, history=None):
        """Create hyperparameter search resutls.

        Parameters
        ----------
        best_model
            best model found during parameter search
        score
            best score achieved
        history
            object containing all hyperparameter search trials
        """
        self.model = best_model
        self.score = best_score
        self.history = history


class Hyperopt(ModelSpaceFunctor):
    """Pipeline step for executing hyperparameter search with Hyperopt library
    """
    def __init__(self, score_step_fn, max_evals=100, reverse_score=True):
        """Hyperparameter search with Hyperopt
        
        Parameters
        ----------
        score_step_fn: callable
            Funtion that will train given model and return resulting score.
            Can be a CV or Validation step.

        max_evals: int
            Limits the maximum number of search steps

        reverse_score: boolean
            As Hyperopt search parameters that minimize the score sometimes
            it's convenient to reverse it so the optimization proceeds in a
            correct way. For example, when using ROC AUC scorer setting this
            parameter to True will result in sound optimization problem
            formulation.
        """
        self._log = logging.getLogger(self.__class__.__name__)
        
        def score_wrapper(*args, **kwargs):
            """Wrap score function so it returns appropriate format for
            hyperopt"""
            score_result = score_step_fn(*args, **kwargs)
            if reverse_score:
                score = 1 - score_result.score 
            else:
                score = score_result.score

            return {
                'loss': score,
                'status': STATUS_OK,
                'model': score_result.model
                }

        self._score_step_fn = score_wrapper

        self._max_evals = max_evals
        self._reverse_score = reverse_score

    def __call__(self, pipeline_data, context):
        model, hparam_space = context
        self._log.info(hparam_space)
        if not hparam_space:
            self._log.warn((f"Skipping hyperopt step for model {model}. No "
                            "parameter templats found"))
            return HyperparameterSearchResult(model, 0, None)

        trials = Trials()
        self._log.info(f"Running hyperparameter optimization for {model}")

        score = partial(self._score_step_fn, pipeline_data, context)

        fmin(score,
             space=hparam_space,
             algo=tpe.suggest,
             trials=trials,
             max_evals=self._max_evals)

        if self._reverse_score:
            self._log.info("Reversing best score bask to original form as reverse_score=True")
            best_score = 1 - sorted(trials.losses())[0]
        else:
            best_score = sorted(trials.losses())[0]

        best = trials.best_trial['result']['model']
        result = HyperparameterSearchResult(best,
                                            best_score,
                                            trials)
        return result



"""Pipeline steps for feature selection"""
import logging
from sklearn.feature_selection import RFE
from automl.pipeline import PipelineData

import numpy as np


def softmax(x):
    return np.exp(x)/np.sum(np.exp(x))


class FeatureSelector:
    """Class for feature selection step in pipeline. This selector
    provides selection strategy based on a single estimator. It means
    that FeatureSelector must be used after ChooseBest(1)

    Parametrs
    ---------
    max_features: int
        Desired numbers of features
    """
    def __init__(self, max_features):
        self._log = logging.getLogger(self.__class__.__name__)
        self.max_features = max_features

    def __call__(self, pipeline_data, context):
        """
        Parametrs
        ---------
        pipeline_data : PipelineData
            Data passed between PipelineStep in pipeline

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        PipelineData
            PipelineData.dataset contains changed dataset
        """
        if len(pipeline_data.return_val) > 1:
            raise ValueError(("Recurcive Feature Selector must be used with"
                              " ChooseBest(1)"))

        data_len = pipeline_data.dataset.data.shape[1]
        mask = np.array([True for _ in range(0, data_len)])
        model = pipeline_data.return_val[0].model
        # TODO this really should not be used with more than one model
        # Use ChooseBest(1)
        # For seceral models use VotingFeatureSelector
        # TODO: CV scorer in hyperopt does not fit models ???       


        if hasattr(model, "coef_"):
            f_score = [abs(coef) for coef in model.coef_]
        elif hasattr(model, "feature_importances_",):
            f_score = [abs(feature_importances) for feature_importances in model.feature_importances_]
        else:
            f_score = None

        if f_score is not None:
            if pipeline_data.dataset.data.shape[1] > self.max_features:
                threshold = sorted(f_score)[-self.max_features]
                mask = np.array([score >= threshold for score in f_score])
                self._log.info((f"Removing {sum(mask)} features for model"
                                f" {model.__class__.__name__}"))
        else:
            self._log.warn((f"Model {model.__class__.__name__} is not"
                            " supported by FeatureSelector"))

        pipeline_data.dataset.data = pipeline_data.dataset.data.compress(
            mask,
            axis=1
        )
        #pipeline_data.dataset.meta = [feature for feature, informative in zip(pipeline_data.dataset.meta, mask) if informative]
        meta = []
        for feature, informative in zip(pipeline_data.dataset.meta, mask):
            if informative:
                meta.append(feature)
        pipeline_data.dataset.meta = meta
        return PipelineData(pipeline_data.dataset, pipeline_data.return_val)


class RecursiveFeatureSelector:
    """Class for feature selection step in pipeline. This selector
    provides recurcive strategy based on a single estimator. It means
    that FeatureSelector must be used after ChooseBest(1)

    Parameters
    ----------
    n_features_to_select : int or None (default=None)
        The number of features to select. If None, half of the features are selected.

    step : int or float, optional (default=1)
        If greater than or equal to 1, then step corresponds to
        the (integer) number of features to remove at each iteration.
        If within (0.0, 1.0), then step corresponds to the percentage
        (rounded down) of features to remove at each iteration.

    verbose : int, default=0
        Controls verbosity of output.

    See also
    --------
    http://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFE.html
    """
    def __init__(self, n_features_to_select=None, step=1, verbose=0):
        self._log = logging.getLogger(self.__class__.__name__)
        self.n_features_to_select = n_features_to_select
        self.step = step
        self.verbose = verbose

    def __call__(self, pipeline_data, context):
        """
        Parametrs
        ---------
        pipeline_data : PipelineData
            Data passed between PipelineStep in pipeline

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        PipelineData
            PipelineData.dataset contains changed dataset
        """
        if len(pipeline_data.return_val) > 1:
            raise ValueError(("Recurcive Feature Selector must be used with"
                              "ChooseBest(1)"))

        model = pipeline_data.return_val[0].model

        if hasattr(model, "coef_") or hasattr(model, "feature_importances_",):
            selector = RFE(
                model,
                self.n_features_to_select,
                self.step,
                self.verbose
            )

            assert(list(selector.get_params()['estimator'].get_params().values()) == list(model.get_params().values()))
            pipeline_data.dataset.data = selector.fit_transform(
                pipeline_data.dataset.data,
                pipeline_data.dataset.target
            )

            informative_features = zip(pipeline_data.dataset.meta,
                                       selector.get_support())
            meta = []
            for feature, informative in informative_features:
                if informative:
                    meta.append(feature)
            pipeline_data.dataset.meta = meta
        else:
            self._log.warn((f"Estimator {model.__class__.__name__} must have"
                            " coef_ or feature_importances_ attribute"))

        return PipelineData(pipeline_data.dataset, pipeline_data.return_val)


class VotingFeatureSelector:
    """Class for feature selection step in pipeline. This selector
    provides selection strategy based on several estimator. It means
    that FeatureSelector can be used with ChooseBest(n), n>1. All
    model that have feature_importances_ or coeff_ attribute have
    contribution to selection

    Parametrs
    ---------
    feature_to_select : int
        Desired numbers of features

    reverse_score : boolean
        If True, it means that bigger value of score is better,
        if False, it means that smaller value of score is worse
    """

    def __init__(self, feature_to_select, reverse_score=False):
        self._log = logging.getLogger(self.__class__.__name__)
        self.feature_to_select = feature_to_select
        self.reverse_score = reverse_score

    def __call__(self, pipeline_data, context):
        """
        Parametrs
        ---------
        pipeline_data : PipelineData
            Data passed between PipelineStep in pipeline

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        PipelineData
            PipelineData.dataset contains changed dataset
        """
        vote = []
        model_scores = []

        for value in pipeline_data.return_val:
            model = value.model

            if hasattr(model, "coef_"):
                f_score = np.array([abs(coef) for coef in model.coef_])
            elif hasattr(model, "feature_importances_",):
                f_score = np.array([abs(feature_importances)
                                    for feature_importances
                                    in model.feature_importances_])
            else:
                f_score = None

            if f_score is not None:
                vote.append(f_score)
                model_scores.append(value.score)

        model_scores = softmax(np.array(model_scores))

        if not vote:
            raise ValueError(("VotingFeatureSelector is needed at least one"
                              "model with feature_importances_ or coef_"
                              "attribute"))
        else:
            weights = np.zeros(vote[0].shape[0])

        if self.reverse_score:
            for f_score, model_score in zip(vote, model_scores):
                weights = weights + softmax(np.array(f_score)) * \
                        (1 - model_score)
        else:
            for f_score, model_score in zip(vote, model_scores):
                weights = weights + softmax(np.array(f_score)) * model_score

        threshold = sorted(weights)[-self.feature_to_select]
        mask = [score >= threshold for score in weights]

        if sum(mask):
            pipeline_data.dataset.data = pipeline_data.dataset.data.compress(
                mask,
                axis=1
            )
            #pipeline_data.dataset.meta = [feature for feature, informative in zip(pipeline_data.dataset.meta, mask) if informative]
            meta = []
            for feature, informative in zip(pipeline_data.dataset.meta, mask):
                if informative:
                    meta.append(feature)
            pipeline_data.dataset.meta = meta

        return PipelineData(pipeline_data.dataset, pipeline_data.return_val)


class CorrelatedFeatureSelector:
    def __init__(self, max_correlation):
        self._log = logging.getLogger(self.__class__.__name__)
        self.max_correlation = max_correlation

    def __call__(self, pipeline_data, context):
        data = pipeline_data.dataset.data
        meta = pipeline_data.dataset.meta
        orig_feature_num = data.shape[1]
        
        
        correlation_matrix = np.corrcoef(data.T)
        matrix_mask = np.absolute(correlation_matrix) > self.max_correlation
        pairs_of_indices = np.array(np.nonzero(matrix_mask)).T
        candidates = []
        for pair in pairs_of_indices:
            if pair[0]!=pair[1]:
                candidates.append(tuple(sorted(pair)))  #collect all pairs of feature indices which feature correlation bigger than max correlation parameter without pairs received by correlation with itself 
        candidates = list(set(candidates)) #delete pairs received by permutation of indices

        indices_to_delete = set([pair[0] for pair in candidates])
        mask = [i not in indices_to_delete for i in range(0, orig_feature_num)]#list of booleans. True if feature is not deleted by generator
        
        pipeline_data.dataset.data = data.compress(mask, axis=1)
        
        new_meta = []
        informative_features = zip(meta, mask) 
        for feature, informative in informative_features:
            if informative:
                new_meta.append(feature)
        new_meta = [feature for feature, informative in zip(meta, mask)
                    if informative]
        
        pipeline_data.dataset.meta = new_meta
        
        return PipelineData(pipeline_data.dataset, pipeline_data.return_val)




class Expression:
    def __add__(self, other):
        return AddExpression(self, other)
    
    def __sub__(self, other):
        return SubExpression(self, other)

    def __mul__(self, other):
        return MulExpression(self, other)
    
    def __truediv__(self, other):
        return DivExpression(self, other)

class Atom(Expression):
    def __init__(self, index):
        self.index = index

    def eval(self, data):
        return data[:, self.index]

class AddExpression(Expression):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def eval(self, data):
        return self.a.eval(data) + self.b.eval(data)

class SubExpression(Expression):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def eval(self, data):
        return self.a.eval(data) - self.b.eval(data)

class MulExpression(Expression):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def eval(self, data):
        return self.a.eval(data) * self.b.eval(data)

class DivExpression(Expression):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def eval(self, data):
        return self.a.eval(data) / self.b.eval(data)


"""Hyperopt templates for different models"""

# forked from hyperopt/hyperopt-sklearn

from functools import partial

import numpy as np
from hyperopt import hp
from hyperopt.pyll import scope

import sklearn.discriminant_analysis
import sklearn.ensemble
import sklearn.feature_extraction.text
import sklearn.preprocessing
import sklearn.svm
import sklearn.tree

# Optional dependencies
try:
    import xgboost
except ImportError:
    xgboost = None


def default_name_func(name):
    return name

##############################
##==== Global variables ====##
##############################
_svm_default_cache_size = 512


###############################################
##==== Various hyperparameter generators ====##
###############################################
def hp_bool(name):
    return hp.choice(name, [False, True])


def _svm_gamma(name, n_features=1):
    '''Generator of default gamma values for SVMs.
    This setting is based on the following rationales:
    1.  The gamma hyperparameter is basically an amplifier for the
        original dot product or l2 norm.
    2.  The original dot product or l2 norm shall be normalized by
        the number of features first.
    '''
    # -- making these non-conditional variables
    #    probably helps the GP algorithm generalize
    # assert n_features >= 1
    return hp.loguniform(name,
                         np.log(1. / n_features * 1e-3),
                         np.log(1. / n_features * 1e3))


def _svm_degree(name):
    return hp.quniform(name, 1.5, 6.5, 1)


def _svm_max_iter(name):
    return hp.qloguniform(name, np.log(1e7), np.log(1e9), 1)


def _svm_C(name):
    return hp.loguniform(name, np.log(1e-5), np.log(1e5))


def _svm_tol(name):
    return hp.loguniform(name, np.log(1e-5), np.log(1e-2))


def _svm_int_scaling(name):
    return hp.loguniform(name, np.log(1e-1), np.log(1e1))


def _svm_epsilon(name):
    return hp.loguniform(name, np.log(1e-3), np.log(1e3))


def _svm_loss_penalty_dual(name):
    """
    The combination of penalty='l1' and loss='hinge' is not supported
    penalty='l2' and loss='hinge' is only supported when dual='true'
    penalty='l1' is only supported when dual='false'.
    """
    return hp.choice(
        name, [('hinge', 'l2', True), ('squared_hinge', 'l2', True),
               ('squared_hinge', 'l1', False), ('squared_hinge', 'l2', False)])


def _knn_metric_p(name, sparse_data=False, metric=None, p=None):
    if sparse_data:
        return ('euclidean', 2)
    elif metric == 'euclidean':
        return (metric, 2)
    elif metric == 'manhattan':
        return (metric, 1)
    elif metric == 'chebyshev':
        return (metric, 0)
    elif metric == 'minkowski':
        assert p is not None
        return (metric, p)
    elif metric is None:
        return hp.pchoice(name, [
            (0.55, ('euclidean', 2)),
            (0.15, ('manhattan', 1)),
            (0.15, ('chebyshev', 0)),
            (0.15, ('minkowski', _knn_p(name + '.p'))),
        ])
    else:
        return (metric, p)  # undefined, simply return user input.


def _knn_p(name):
    return hp.quniform(name, 2.5, 5.5, 1)


def _knn_neighbors(name):
    return scope.int(hp.qloguniform(name, np.log(0.5), np.log(50.5), 1))


def _knn_weights(name):
    return hp.choice(name, ['uniform', 'distance'])


def _trees_n_estimators(name):
    return scope.int(hp.qloguniform(name, np.log(9.5), np.log(3000.5), 1))


def _trees_criterion(name):
    return hp.choice(name, ['gini', 'entropy'])


def _trees_max_features(name):
    return hp.pchoice(
        name,
        [
            (0.2, 'sqrt'),  # most common choice.
            (0.1, 'log2'),  # less common choice.
            (0.1, None),  # all features, less common choice.
            (0.6, hp.uniform(name + '.frac', 0., 1.))
        ])


def _trees_max_depth(name):
    return hp.pchoice(
        name,
        [
            (0.7, None),  # most common choice.
            # Try some shallow trees.
            (0.1, 2),
            (0.1, 3),
            (0.1, 4),
        ])


def _trees_min_samples_split(name):
    return 2


def _trees_min_samples_leaf(name):
    return hp.choice(
        name,
        [
            1,  # most common choice.
            scope.int(
                hp.qloguniform(name + '.gt1', np.log(1.5), np.log(50.5), 1))
        ])


def _trees_bootstrap(name):
    return hp.choice(name, [True, False])


def _boosting_n_estimators(name):
    return scope.int(hp.qloguniform(name, np.log(10.5), np.log(1000.5), 1))


def _ada_boost_learning_rate(name):
    return hp.lognormal(name, np.log(0.01), np.log(10.0))


def _ada_boost_loss(name):
    return hp.choice(name, ['linear', 'square', 'exponential'])


def _ada_boost_algo(name):
    return hp.choice(name, ['SAMME', 'SAMME.R'])


def _grad_boosting_reg_loss_alpha(name):
    return hp.choice(name, [('ls', 0.9), ('lad', 0.9),
                            ('huber', hp.uniform(name + '.alpha', 0.85, 0.95)),
                            ('quantile', 0.5)])


def _grad_boosting_clf_loss(name):
    return hp.choice(name, ['deviance', 'exponential'])


def _grad_boosting_learning_rate(name):
    return hp.lognormal(name, np.log(0.01), np.log(10.0))


def _grad_boosting_subsample(name):
    return hp.pchoice(
        name,
        [
            (0.2, 1.0),  # default choice.
            (0.8, hp.uniform(name + '.sgb', 0.5, 1.0)
             )  # stochastic grad boosting.
        ])


def _sgd_penalty(name):
    return hp.pchoice(name, [(0.40, 'l2'), (0.35, 'l1'), (0.25, 'elasticnet')])


def _sgd_alpha(name):
    return hp.loguniform(name, np.log(1e-6), np.log(1e-1))


def _sgd_l1_ratio(name):
    return hp.uniform(name, 0, 1)


def _sgd_epsilon(name):
    return hp.loguniform(name, np.log(1e-7), np.log(1))


def _sgdc_learning_rate(name):
    return hp.pchoice(name, [(0.50, 'optimal'), (0.25, 'invscaling'),
                             (0.25, 'constant')])


def _sgdr_learning_rate(name):
    return hp.pchoice(name, [(0.50, 'invscaling'), (0.25, 'optimal'),
                             (0.25, 'constant')])


def _sgd_eta0(name):
    return hp.loguniform(name, np.log(1e-5), np.log(1e-1))


def _sgd_power_t(name):
    return hp.uniform(name, 0, 1)


def _random_state(name, random_state):
    if random_state is None:
        return hp.randint(name, 5)
    else:
        return random_state


def _class_weight(name):
    return hp.choice(name, [None, 'balanced'])


##############################################
##==== SVM hyperparameters search space ====##
##############################################
def _svm_hp_space(kernel,
                  n_features=1,
                  C=None,
                  gamma=None,
                  coef0=None,
                  degree=None,
                  shrinking=None,
                  tol=None,
                  max_iter=None,
                  verbose=False,
                  cache_size=_svm_default_cache_size):
    '''Generate SVM hyperparamters search space
    '''
    if kernel in ['linear', 'rbf', 'sigmoid']:
        degree_ = 1
    else:
        degree_ = (_svm_degree('degree')
                   if degree is None else degree)
    if kernel in ['linear']:
        gamma_ = 'auto'
    else:
        gamma_ = (_svm_gamma('gamma', n_features=1)
                  if gamma is None else gamma)
        gamma_ /= n_features  # make gamma independent of n_features.
    if kernel in ['linear', 'rbf']:
        coef0_ = 0.0
    elif coef0 is None:
        if kernel == 'poly':
            coef0_ = hp.pchoice(
                'coef0',
                [(0.3, 0),
                 (0.7, gamma_ * hp.uniform('coef0val', 0., 10.))])
        elif kernel == 'sigmoid':
            coef0_ = hp.pchoice(
                'coef0',
                [(0.3, 0),
                 (0.7, gamma_ * hp.uniform('coef0val', -10., 10.))])
        else:
            pass
    else:
        coef0_ = coef0

    hp_space = dict(
        kernel=kernel,
        C=_svm_C('C') if C is None else C,
        gamma=gamma_,
        coef0=coef0_,
        degree=degree_,
        shrinking=(hp_bool('shrinking')
                   if shrinking is None else shrinking),
        tol=_svm_tol('tol') if tol is None else tol,
        max_iter=(_svm_max_iter('maxiter')
                  if max_iter is None else max_iter),
        verbose=verbose,
        cache_size=cache_size)
    return hp_space


def _svc_hp_space(random_state=None, probability=False):
    '''Generate SVC specific hyperparamters
    '''
    hp_space = dict(
        random_state=_random_state('rstate', random_state),
        probability=probability)
    return hp_space


def _svr_hp_space(epsilon=None):
    '''Generate SVR specific hyperparamters
    '''
    hp_space = {}
    hp_space['epsilon'] = (_svm_epsilon('epsilon')
                           if epsilon is None else epsilon)
    return hp_space


#########################################
##==== SVM classifier constructors ====##
#########################################
def svc_kernel_hp_space(kernel,
                        random_state=None,
                        probability=False,
                        **kwargs):
    """
    Return a hyperparamter template that will construct
    a sklearn.svm.SVC model with a user specified kernel.
    Supported kernels: linear, rbf, poly and sigmoid
    """

    hp_space = _svm_hp_space(kernel=kernel, **kwargs)
    hp_space.update(_svc_hp_space(random_state, probability))
    return hp_space



########################################
##==== SVM regressor constructors ====##
########################################
def svr_kernel_hp_space(kernel, epsilon=None, **kwargs):
    """
    Return a hyperparamter template that will construct
    a sklearn.svm.SVR model with a user specified kernel.
    Supported kernels: linear, rbf, poly and sigmoid
    """

    hp_space = _svm_hp_space(kernel=kernel, **kwargs)
    hp_space.update(_svr_hp_space(epsilon))
    return hp_space


##############################################
##==== KNN hyperparameters search space ====##
##############################################
def knn_hp_space(sparse_data=False,
                 n_neighbors=None,
                 weights=None,
                 algorithm='auto',
                 leaf_size=30,
                 metric=None,
                 p=None,
                 metric_params=None,
                 n_jobs=1):
    '''Generate KNN hyperparameters search space
    '''
    metric_p = _knn_metric_p('metric_p', sparse_data, metric, p)
    hp_space = dict(
        n_neighbors=(_knn_neighbors('neighbors')
                     if n_neighbors is None else n_neighbors),
        weights=(_knn_weights('weights')
                 if weights is None else weights),
        algorithm=algorithm,
        leaf_size=leaf_size,
        metric=metric_p[0] if metric is None else metric,
        p=metric_p[1] if p is None else p,
        metric_params=metric_params,
        n_jobs=n_jobs)
    return hp_space


####################################################################
##==== Random forest/extra trees hyperparameters search space ====##
####################################################################
def trees_hp_space(n_estimators=None,
                   max_features=None,
                   max_depth=None,
                   min_samples_split=None,
                   min_samples_leaf=None,
                   bootstrap=None,
                   oob_score=False,
                   n_jobs=1,
                   random_state=None,
                   verbose=False):
    '''Generate trees ensemble hyperparameters search space
    '''
    hp_space = dict(
        n_estimators=(_trees_n_estimators('n_estimators')
                      if n_estimators is None else n_estimators),
        max_features=(_trees_max_features('max_features')
                      if max_features is None else max_features),
        max_depth=(_trees_max_depth('max_depth')
                   if max_depth is None else max_depth),
        min_samples_split=(_trees_min_samples_split(
            'min_samples_split') if min_samples_split is None else
                           min_samples_split),
        min_samples_leaf=(_trees_min_samples_leaf(
            'min_samples_leaf')
                          if min_samples_leaf is None else min_samples_leaf),
        bootstrap=(_trees_bootstrap('bootstrap')
                   if bootstrap is None else bootstrap),
        oob_score=oob_score,
        n_jobs=n_jobs,
        random_state=_random_state('rstate', random_state),
        verbose=verbose,
    )
    return hp_space


#############################################################
##==== Random forest classifier/regressor constructors ====##
#############################################################
def random_forest_hp_space(criterion='gini', **kwargs):
    """"Return a hyperparameter template for RandomForest model.

    Parameters
    ----------
    criterion: str
        'gini' or 'entropy' and 'mse' for classification
    """

    hp_space = trees_hp_space(**kwargs)
    hp_space['criterion'] = criterion
    return hp_space


###################################################
##==== AdaBoost hyperparameters search space ====##
###################################################
def ada_boost_hp_space(base_estimator=None,
                       n_estimators=None,
                       learning_rate=None,
                       random_state=None):
    '''Generate AdaBoost hyperparameters search space
    '''
    hp_space = dict(
        base_estimator=base_estimator,
        n_estimators=(_boosting_n_estimators('n_estimators')
                      if n_estimators is None else n_estimators),
        learning_rate=(_ada_boost_learning_rate('learning_rate')
                       if learning_rate is None else learning_rate),
        random_state=_random_state('rstate', random_state))
    return hp_space


###########################################################
##==== GradientBoosting hyperparameters search space ====##
###########################################################
def grad_boosting_hp_space(learning_rate=None,
                           n_estimators=None,
                           subsample=None,
                           min_samples_split=None,
                           min_samples_leaf=None,
                           max_depth=None,
                           init=None,
                           random_state=None,
                           max_features=None,
                           verbose=0,
                           max_leaf_nodes=None,
                           warm_start=False,
                           presort='auto'):
    '''Generate GradientBoosting hyperparameters search space
    '''
    hp_space = dict(
        learning_rate=(_grad_boosting_learning_rate('learning_rate')
                        if learning_rate is None else learning_rate),
        n_estimators=(_boosting_n_estimators('n_estimators')
                        if n_estimators is None else n_estimators),
        subsample=(_grad_boosting_subsample('subsample')
                        if subsample is None else subsample),
        min_samples_split=(_trees_min_samples_split('min_samples_split')
                        if min_samples_split is None else min_samples_split),
        min_samples_leaf=(_trees_min_samples_leaf('min_samples_leaf')
                        if min_samples_leaf is None else min_samples_leaf),
        max_depth=(_trees_max_depth('max_depth')
                        if max_depth is None else max_depth),
        init=init,
        random_state=_random_state('rstate', random_state),
        max_features=(_trees_max_features('max_features')
                      if max_features is None else max_features),
        warm_start=warm_start,
        presort=presort)
    return hp_space


###################################################
##==== XGBoost hyperparameters search space ====##
###################################################


def _xgboost_max_depth(name):
    return scope.int(hp.uniform(name, 1, 11))


def _xgboost_learning_rate(name):
    return hp.loguniform(name, np.log(0.0001), np.log(0.5)) - 0.0001


def _xgboost_n_estimators(name):
    return scope.int(hp.quniform(name, 100, 6000, 200))


def _xgboost_gamma(name):
    return hp.loguniform(name, np.log(0.0001), np.log(5)) - 0.0001


def _xgboost_min_child_weight(name):
    return scope.int(hp.loguniform(name, np.log(1), np.log(100)))


def _xgboost_subsample(name):
    return hp.uniform(name, 0.5, 1)


def _xgboost_colsample_bytree(name):
    return hp.uniform(name, 0.5, 1)


def _xgboost_colsample_bylevel(name):
    return hp.uniform(name, 0.5, 1)


def _xgboost_reg_alpha(name):
    return hp.loguniform(name, np.log(0.0001), np.log(1)) - 0.0001


def _xgboost_reg_lambda(name):
    return hp.loguniform(name, np.log(1), np.log(4))


def xgboost_hp_space(max_depth=None,
                     learning_rate=None,
                     n_estimators=None,
                     gamma=None,
                     min_child_weight=None,
                     max_delta_step=0,
                     subsample=None,
                     colsample_bytree=None,
                     colsample_bylevel=None,
                     reg_alpha=None,
                     reg_lambda=None,
                     scale_pos_weight=1,
                     base_score=0.5,
                     random_state=None):
    '''Generate XGBoost hyperparameters search space
    '''
    hp_space = dict(
        max_depth=(_xgboost_max_depth('max_depth')
                   if max_depth is None else max_depth),
        learning_rate=(_xgboost_learning_rate('learning_rate')
                       if learning_rate is None else learning_rate),
        n_estimators=(_xgboost_n_estimators('n_estimators')
                      if n_estimators is None else n_estimators),
        gamma=(_xgboost_gamma('gamma') if gamma is None else gamma),
        min_child_weight=(_xgboost_min_child_weight(
            'min_child_weight')
                          if min_child_weight is None else min_child_weight),
        max_delta_step=max_delta_step,
        subsample=(_xgboost_subsample('subsample')
                   if subsample is None else subsample),
        colsample_bytree=(_xgboost_colsample_bytree(
            'colsample_bytree')
                          if colsample_bytree is None else colsample_bytree),
        colsample_bylevel=(_xgboost_colsample_bylevel(
            'colsample_bylevel') if colsample_bylevel is None else
                           colsample_bylevel),
        reg_alpha=(_xgboost_reg_alpha('reg_alpha')
                   if reg_alpha is None else reg_alpha),
        reg_lambda=(_xgboost_reg_lambda('reg_lambda')
                    if reg_lambda is None else reg_lambda),
        scale_pos_weight=scale_pos_weight,
        base_score=base_score,
        seed=_random_state('rstate', random_state))
    return hp_space


#################################################
##==== Naive Bayes classifiers constructor ====##
#################################################
def multinomial_nb_hp_space(class_prior=None):
    hp_space = dict(
        alpha=hp.quniform('alpha', 0, 1, 0.001),
        fit_prior=hp_bool('fit_prior'),
        class_prior=class_prior)
    return hp_space


###########################################
##==== Passive-aggressive classifier ====##
###########################################
def passive_aggressive_hp_space(loss=None,
                                C=None,
                                fit_intercept=False,
                                n_iter=None,
                                n_jobs=1,
                                random_state=None,
                                verbose=False):
    hp_space = dict(
        loss=hp.choice('loss', ['hinge', 'squared_hinge'])
        if loss is None else loss,
        C=hp.lognormal('learning_rate', np.log(0.01), np.log(10))
        if C is None else C,
        fit_intercept=fit_intercept,
        n_iter=scope.int(
            hp.qloguniform('n_iter', np.log(1), np.log(1000), q=1))
        if n_iter is None else n_iter,
        n_jobs=n_jobs,
        random_state=_random_state('rstate', random_state),
        verbose=verbose)

    return hp_space


###############################################
##==== Discriminant analysis classifiers ====##
###############################################
def linear_discriminant_analysis_hp_space(solver=None,
                                          shrinkage=None,
                                          priors=None,
                                          n_components=None,
                                          store_covariance=False,
                                          tol=0.00001):

    solver_shrinkage = hp.choice('solver_shrinkage_dual',
                                 [('svd', None), ('lsqr', None),
                                  ('lsqr', 'auto'), ('eigen', None),
                                  ('eigen', 'auto')])

    rval = dict(
        solver=solver_shrinkage[0] if solver is None else solver,
        shrinkage=solver_shrinkage[1] if shrinkage is None else shrinkage,
        priors=priors,
        n_components=4 * scope.int(
            hp.qloguniform(
                'n_components', low=np.log(0.51), high=np.log(30.5), q=1.0))
        if n_components is None else n_components,
        store_covariance=store_covariance,
        tol=tol)
    return rval


def quadratic_discriminant_analysis_hp_space(reg_param=None, priors=None):

    rval = dict(
        reg_param=hp.uniform('reg_param', 0.0, 1.0)
        if reg_param is None else 0.0,
        priors=priors)
    return rval


###############################################
##==== Various preprocessor constructors ====##
###############################################
def pca_hp_space(n_components=None, whiten=None, copy=True):
    rval = dict(
        # -- qloguniform is missing a "scale" parameter so we
        #    lower the "high" parameter and multiply by 4 out front
        n_components=4 * scope.int(
            hp.qloguniform(
                'n_components', low=np.log(0.51), high=np.log(30.5), q=1.0))
        if n_components is None else n_components,
        # n_components=(hp.uniform(name + '.n_components', 0, 1)
        #               if n_components is None else n_components),
        whiten=hp_bool('whiten') if whiten is None else whiten,
        copy=copy,
    )
    return rval


def standard_scaler(with_mean=None, with_std=None):
    rval = dict(
        with_mean=hp_bool('with_mean') if with_mean is None else with_mean,
        with_std=hp_bool('with_std') if with_std is None else with_std,
    )
    return rval


def ts_lagselector_hp_space(lower_lags=1, upper_lags=1):
    rval = dict(lag_size=scope.int(
        hp.quniform('lags', lower_lags - .5, upper_lags + .5, 1)))
    return rval


def bernoulli_rbm_hp_space(n_components=None,
                           learning_rate=None,
                           batch_size=None,
                           n_iter=None,
                           verbose=False,
                           random_state=None):

    rval = dict(
        n_components=scope.int(
            hp.qloguniform(
                'n_components', low=np.log(0.51), high=np.log(999.5), q=1.0))
        if n_components is None else n_components,
        learning_rate=hp.lognormal(
            'learning_rate',
            np.log(0.01),
            np.log(10),
        ) if learning_rate is None else learning_rate,
        batch_size=scope.int(
            hp.qloguniform(
                '.batch_size',
                np.log(1),
                np.log(100),
                q=1,
            )) if batch_size is None else batch_size,
        n_iter=scope.int(
            hp.qloguniform(
                'n_iter',
                np.log(1),
                np.log(1000),  # -- max sweeps over the *whole* train set
                q=1,
            )) if n_iter is None else n_iter,
        verbose=verbose,
        random_state=_random_state('rstate', random_state),
    )
    return rval


def colkmeans_hp_space(n_clusters=None,
                       init=None,
                       n_init=None,
                       max_iter=None,
                       tol=None,
                       precompute_distances=True,
                       verbose=0,
                       random_state=None,
                       copy_x=True,
                       n_jobs=1):
    rval = dict(
        n_clusters=scope.int(
            hp.qloguniform(
                'n_clusters', low=np.log(1.51), high=np.log(19.5), q=1.0))
        if n_clusters is None else n_clusters,
        init=hp.choice(
            'init',
            ['k-means++', 'random'],
        ) if init is None else init,
        n_init=hp.choice(
            'n_init',
            [1, 2, 10, 20],
        ) if n_init is None else n_init,
        max_iter=scope.int(
            hp.qlognormal(
                'max_iter',
                np.log(300),
                np.log(10),
                q=1,
            )) if max_iter is None else max_iter,
        tol=hp.lognormal(
            'tol',
            np.log(0.0001),
            np.log(10),
        ) if tol is None else tol,
        precompute_distances=precompute_distances,
        verbose=verbose,
        random_state=random_state,
        copy_x=copy_x,
        n_jobs=n_jobs,
    )
    return rval

def lgbm_hp_space(**kwargs):
    space = {
        'n_estimators': scope.int(hp.quniform('n_estimators', 10, 700, 1)),
        'num_leaves': scope.int(hp.quniform ('num_leaves', 10, 200, 1)),
        'feature_fraction': hp.uniform('feature_fraction', 0.75, 1.0),
        'bagging_fraction': hp.uniform('bagging_fraction', 0.75, 1.0),
        'learning_rate': hp.loguniform('learning_rate', -5.0, -2.3),
        'max_bin': scope.int(hp.quniform('max_bin', 64, 512, 1)),
        'bagging_freq': scope.int(hp.quniform('bagging_freq', 1, 5, 1)),
        'lambda_l1': hp.uniform('lambda_l1', 0, 10),
        'lambda_l2': hp.uniform('lambda_l2', 0, 10),
        **kwargs
       }

    return space
# -- flake8 eofk



"""Dataset loading from external sources"""
import logging
from urllib.parse import urlparse

import pandas as pd

from automl.pipeline import PipelineData


class DatasetLoader:
    """Data loading for common formats"""

    def __init__(self, **kwargs):
        self._log = logging.getLogger(self.__class__.__name__)
        self._loader_map = {
            'csv': self._load_csv
        }
        self._loader_args = kwargs

    def __call__(self, resource_uri, pipeline_context=None):
        """Load resource of supported type.

        You can use pandas options for most formats:
        >>> default_dataset_loader("csv:///data.csv", delimeter=";")

        
        Parameters
        ----------
        resource_uri: str
            URI in form of `format:///path/to/resource`
        kwargs:
            keyword arguments are passed to concrete loader implementation


        Returns
        -------
        dataframe: pandas.DataFrame
            resulting data
        """

        uri = urlparse(resource_uri)
        if uri.scheme == '':
            raise ValueError("Please provide format for loading")
        if uri.scheme not in self._loader_map:
            raise ValueError(f"{uri.scheme} is not supported")

        return PipelineData(self._loader_map[uri.scheme](uri.path, **self._loader_args))

    def _load_csv(self, csv_path, **kwargs):
        return pd.read_csv(csv_path, **self._loader_args)

    def register_custom_loader(self, format_name, loader_func):
        """Registers custom data loader function.
        
        Parameters
        ----------
        format_name: str
            format name to use as a scheme prefix in URI
        loader_func: callable
            callable that performs data loading
        """
        if not callable(loader_func):
            raise ValueError("loader_func must be callable")
        self._loader_map[format_name] = loader_func



"""Dataset abstraction"""
import numpy as np

from automl.expression import Atom

class Dataset:
    """AutoML Dataset. This class is the main input object that needs to be
    passed into the Pipeline
    """

    def __init__(self, data, target):
        """Create Dataset.

        Parameters
        ----------
        data: pandas.DataFrame or numpy.array
            training data
        target: pandas.DataFrame or numpy.array
            target variable
        """
        if hasattr(data, 'columns'):
            self.meta = []
            for i, col in enumerate(data.columns):
                self.meta.append({"name": col,
                                  "history": Atom(i),
                                  "depth": 1})

        else:
            try:
                self.meta = [{"name": f"base_feature_{i}",
                              "history": Atom(i),
                              "depth": 1} for i in range(0, data.shape[1])]
            except Exception:
                raise ValueError("Can't construct dataset. Please use two dimensinal numpy arrays or pandas dataframes")


        if not isinstance(data, np.ndarray):
            data = np.array(data, dtype='float32')
        else:
            data = data.astype('float32')
        self.data = data
        self.target = target

    @property
    def columns(self):
        """Get column names

        Returns
        -------
        columns: list[str]
        """
        return [m['name'] for m in self.meta]


class DatasetExtractor:
    def __init__(self, target, data_col_filter=None):
        """Extract Dataset from pandas DataFrame

        Parameters
        ----------
        target: str or int
            target column name
        data_col_filter: list of str, str, or callable
            filter data columns if needed. Can be function that accepts input
            dataset and return a column list
        """
        self._target = target

        self._data_col_filter = data_col_filter

    def __call__(self, x, context):
        data = x.drop(self._target, axis=1)
        if self._data_col_filter is not None:
            if callable(self._data_col_filter):
                col_list = self._data_col_filter(x)
            else:
                col_list = self._data_col_filter

            data = data[col_list]

        return Dataset(data, x[self._target])



from setuptools import setup, find_packages
from automl.constants import *

setup(name='onepanel-automl',
      version=AUTOML_VERSION,
      description='Automated Machine Learning Framework',
      author='Onepanel, Inc.',
      author_email='info@onepanel.io',
      packages=find_packages(),
      install_requires=[
          'numpy>=1.13.3',
          'pandas>=0.20.3',
          'scikit-learn>=0.19.0',
          'tqdm',
          'hyperopt>=0.1',
          'networkx==1.11'
          ],
      license='License :: Other/Proprietary License',
      classifiers=['Topic :: Scientific/Engineering :: Artificial Intelligence',
                   'Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 3.6'])



"""Pipeline steps for feature generation"""

import logging
import itertools as it
from functools import partial
import random
import numpy as np
import operator
from sklearn.preprocessing import PolynomialFeatures

from automl.expression import Atom
    

class PolynomialFeatureGenerator:
    """Generate polynomial and interaction features for dataset in PipelineData

    Generate a new feature matrix from current matrix in pipeline
    consisting of all polynomial combinations of the features with
    degree less than or equal to the specified degree. For example,
    if an input sample is two dimensional and of the form [a, b],
    the degree-2 polynomial features are [1, a, b, a^2, ab, b^2].

    Parameters
    ----------
    max_degree : integer
        The maximum degree of the polynomial features. Default = 2.

    Notes
    -----
    Be aware that the number of features in the output array scales
    polynomially in the number of features of the input array, and
    exponentially in the degree. High degrees can cause overfitting.
    """
    def __init__(self, max_degree):
        self._log = logging.getLogger(self.__class__.__name__)
        self.max_degree = max_degree

    def __call__(self, pipeline_data, pipeline_context):
        """
        Parametrs
        ---------
        pipeline_data : PipelineData
            Data passed between PipelineStep in pipeline

        context : PiplineContext
            Global context of pipeline

        Returns
        -------
        PipelineData
            PipelineData containing changed PipelineData.dataset
        """
        data = pipeline_data.dataset.data
        meta = pipeline_data.dataset.meta
        orig_feature_num = pipeline_data.dataset.data.shape[1]

        for degree in range(1, self.max_degree+1):
            # TODO this is unreadable, unwrap
            sets_of_indices = list(set(tuple(sorted(indices)) for indices in it.product(range(0, orig_feature_num), repeat=degree)))

            for indices in sets_of_indices:
                new_feature = np.ones((data.shape[0], 1), dtype='float32')
                history = Atom(0) / Atom(0) #init one as expression

                for index in indices:
                    new_feature = np.reshape(
                        data[:, index],
                        (data.shape[0], 1)
                    ) * new_feature
                    history = meta[index]['history'] * history
                if np.isfinite(new_feature).all():
                    data = np.append(data, new_feature, axis=1)
                    meta.append({
                        "name": "",
                        "history": history 
                    })
                else:
                    pass

        pipeline_data.dataset.data = data
        pipeline_data.dataset.meta = meta
        return pipeline_data


class SklearnFeatureGenerator:
    """ Wrapper for Scikit-Learn Transformers (only for
    sklearn.preprocessing.PolynomialFeatures at this release)

    Parameters
    ----------
    transformer_class: sklearn.preprocessing.PolynomialFeatures

    args:
        arguments are passed to sklearn PolynomialFeatures

    kwargs:
        keyword arguments are passed to sklearn PolynomialFeatures
    """
    def __init__(self, transformer_class, *args, **kwargs):
        self._log = logging.getLogger(self.__class__.__name__)
        self._transformer = transformer_class(*args, **kwargs)

    def __call__(self, pipeline_data, pipeline_context):
        """
        Parameters
        ----------
        pipeline_data : PipelineData
            Data passed between PipelineStep in pipeline

        pipeline_context: automl.pipeline.PipelineContext
            global context of a pipeline

        Returns
        -------
        PipelineData
            PipelineData containing changed PipelineData.dataset
        """

        pipeline_data.dataset.data = self._transformer.fit_transform(pipeline_data.dataset.data)
        return pipeline_data


class FormulaFeatureGenerator:
    """ Generate features for dataset in PipelineData by formula

    Generate a new feature matrix from current matrix in pipeline
    consisting of old features and new features generated from operations
    over old features

    Paramets
    --------
    func_list : list of symbols of functions
        In current version func_list may contain only '+', '-', '*', '/'.

    limit: int
        Total number of expression members

    Attributes
    ----------
    _func_map : dict of generating functions
        The function for generation new features

    used_func : set of of symbols of functions
    """

    def __init__(self, func_list=['+', '-', '*', '/'], limit=1, max_depth=4, addition=0.25):
        self._log = logging.getLogger(self.__class__.__name__)
        self.used_func = set(func_list)
        self._func_map = {
            '+': self._sum,
            '-': self._substract,
            '/': self._divide,
            '*': self._multiply,
        }
        self._limit = limit
        self.max_depth = max_depth
        self.addition = addition

    def _execution(self, dataset, operator):
        X = dataset.data
        first_index, second_index = self._choose_two_index(X)
        x, y = X[:, first_index].reshape(
            X.shape[0], 1), X[:, second_index].reshape(X.shape[0], 1)
        history = operator(dataset.meta[first_index]['history'], dataset.meta[second_index]['history'])
        name = f"({dataset.meta[first_index]['name']}_{operator.__name__}_{dataset.meta[second_index]['name']})"
        depth = dataset.meta[first_index]['depth'] + dataset.meta[first_index]['depth']
        return operator(x, y), history, name, depth

    def _sum(self, dataset):
        """ Generate one new feature by sum of two random features

        Parametrs
        ---------
        dataset : Dataset
            Dataset.data contains data to transform

        Returns
        -------
        np.ndarray shape [n_sample, 1]
            new feature

        history : str
            the history of new feature for reproducible preprocessing
        """
        
        return self._execution(dataset, operator.__add__)

    def _substract(self, dataset):
        """Generate one new feature by substraction of two random features

        Parametrs
        ---------
        dataset : Dataset
            Dataset.data contains data to transform

        Returns
        -------
        np.ndarray shape [n_sample, 1]
            new feature

        history : str
            the history of new feature for reproducible preprocessing
        """
        return self._execution(dataset, operator.__sub__)

    def _divide(self, dataset):
        """ Generate one new feature by division of two random features

        Parametrs
        ---------
        dataset : Dataset
            Dataset.data contains data to transform
            
        Returns
        -------
        np.ndarray shape [n_sample, 1]
            new feature

        history : str
            the history of new feature for reproducible preprocessing
        """
        return self._execution(dataset, operator.__truediv__)

    def _multiply(self, dataset):
        """ Generate one new feature by multiplication of two random features

        Parametrs
        ---------
        dataset : Dataset
            Dataset.data contains data to transform

        Returns
        -------
        np.ndarray shape [n_sample, 1]
            new feature

        history : str
            the history of new feature for reproducible preprocessing
        """
        return self._execution(dataset, operator.__mul__)

    def _choose_two_index(self, X):
        """ Choose two index input data

        Parametrs
        ---------
        X : np.ndarray, shape [n_samples, n_features]
            The data

        Returns
        -------
        int: 
            Index

        int:
            Index 
        """
        return random.randint(0, X.shape[1]-1),\
               random.randint(0, X.shape[1]-1)

    def __call__(self, pipeline_data, pipeline_context):
        """
        Parameters
        ----------
        pipeline_data : automl.pipeline.PipelineData
            Data passed between PipelineStep in pipeline

        pipeline_context : automl.pipeline.PipelineContext
            global context of a pipeline

        limit : int
            Maximum amount of new features. Default = 10

        Returns
        -------
        PipelineData
            PipelineData containing changed PipelineData.dataset
        """
        orig_feature_num = pipeline_data.dataset.data.shape[1]

        for _ in range(0, self._limit):
            new_feature, history, name, depth = self._func_map[random.sample(self.used_func, 1)[0]](pipeline_data.dataset)
            if np.isfinite(new_feature).all() and depth < self.max_depth:
                pipeline_data.dataset.data = np.append(pipeline_data.dataset.data, new_feature, axis=1)

                pipeline_data.dataset.meta.append({
                    "name": name,
                    "history": history,
                    "depth": depth
                })

        self._log.info((f"Generated new features. Old feature number - "
                        f"{orig_feature_num}, new feature number - "
                        f"{pipeline_data.dataset.data.shape[1]}"))
        self.max_depth = self.max_depth + self.addition
        return pipeline_data 

class Preprocessing:
    """ Reproduce all feature transformations that were creating during the
    execution of AutoML pipeline
    """

    def __init__(self):
        self._log = logging.getLogger(self.__class__.__name__)

    def reproduce(self, resulting_dataset, original_dataset):
        """ Parameters
        ----------
        resulting_dataset : Dataset
            Resulting dataset

        original_dataset : Dataset
            Initial dataset that was passed to executor
            
        Returns
        -------
        np.ndarray
            Dataset that was generated by executor
        """
        data = original_dataset.data.astype("float32")
        final_data = np.ones((data.shape[0], 1), dtype='float32')
        for feature in resulting_dataset.meta:
            new_feature = feature['history'].eval(data)
            new_feature = np.reshape(
                new_feature,
                (data.shape[0], 1)
            )
            final_data = np.append(final_data, new_feature, axis=1) 
        return np.delete(final_data.astype('float32'), 0, 1)

PolynomialGenerator = partial(SklearnFeatureGenerator, PolynomialFeatures)
